from __future__ import annotations

from typing import cast

from coredis.response._callbacks import ResponseCallback
from coredis.response._utils import flat_pairs_to_dict
from coredis.typing import (
    AnyStr,
    Dict,
    List,
    Optional,
    ResponseType,
    StringT,
    Tuple,
    TypeGuard,
    Union,
    ValueT,
)


class HScanCallback(
    ResponseCallback[
        List[ResponseType], List[ResponseType], Tuple[int, Dict[AnyStr, AnyStr]]
    ]
):
    def guard(
        self, response: List[ResponseType]
    ) -> TypeGuard[Tuple[StringT, List[AnyStr]]]:
        return isinstance(response[0], (str, bytes)) and isinstance(response[1], list)

    def transform(
        self, response: List[ResponseType], **options: Optional[ValueT]
    ) -> Tuple[int, Dict[AnyStr, AnyStr]]:
        assert self.guard(response)
        cursor, r = response
        return int(cursor), flat_pairs_to_dict(r)


class HRandFieldCallback(
    ResponseCallback[
        Union[AnyStr, List[AnyStr]],
        Union[AnyStr, List[List[AnyStr]]],
        Optional[Union[AnyStr, Tuple[AnyStr, ...], Dict[AnyStr, AnyStr]]],
    ]
):
    def transform(
        self, response: Union[AnyStr, List[AnyStr]], **options: Optional[ValueT]
    ) -> Optional[Union[AnyStr, Tuple[AnyStr, ...], Dict[AnyStr, AnyStr]]]:
        if not response:
            return None
        if options.get("count"):
            assert isinstance(response, list)
            if options.get("withvalues"):
                return flat_pairs_to_dict(response)
            else:
                return tuple(response)
        assert isinstance(response, (str, bytes))
        return response

    def transform_3(
        self, response: Union[AnyStr, List[List[AnyStr]]], **options: Optional[ValueT]
    ) -> Optional[Union[AnyStr, Tuple[AnyStr, ...], Dict[AnyStr, AnyStr]]]:
        if not response:
            return None
        if options.get("count"):
            assert isinstance(response, list)
            if options.get("withvalues"):
                return dict(cast(List[Tuple[AnyStr, AnyStr]], response))
            return tuple(cast(Tuple[AnyStr, AnyStr], response))
        assert isinstance(response, (str, bytes))
        return response


class HGetAllCallback(
    ResponseCallback[List[AnyStr], Dict[AnyStr, AnyStr], Dict[AnyStr, AnyStr]]
):
    def transform(
        self, response: List[AnyStr], **options: Optional[ValueT]
    ) -> Dict[AnyStr, AnyStr]:

        return flat_pairs_to_dict(response) if response else {}

    def transform_3(
        self, response: Dict[AnyStr, AnyStr], **options: Optional[ValueT]
    ) -> Dict[AnyStr, AnyStr]:
        return response



from __future__ import annotations

from typing import cast

from coredis.response._callbacks import ResponseCallback
from coredis.response._utils import flat_pairs_to_dict
from coredis.typing import (
    AnyStr,
    Dict,
    List,
    Optional,
    ResponsePrimitive,
    ResponseType,
    Tuple,
    ValueT,
)


class ModuleInfoCallback(
    ResponseCallback[
        List[List[ResponseType]],
        List[Dict[AnyStr, ResponsePrimitive]],
        Tuple[Dict[AnyStr, ResponsePrimitive], ...],
    ]
):
    def transform(
        self, response: List[List[ResponseType]], **options: Optional[ValueT]
    ) -> Tuple[Dict[AnyStr, ResponsePrimitive], ...]:
        return tuple(
            cast(Dict[AnyStr, ResponsePrimitive], flat_pairs_to_dict(mod))
            for mod in response
        )

    def transform_3(
        self,
        response: List[Dict[AnyStr, ResponsePrimitive]],
        **options: Optional[ValueT],
    ) -> Tuple[Dict[AnyStr, ResponsePrimitive], ...]:
        return tuple(response)



from __future__ import annotations

import asyncio
import sys
from abc import ABC, abstractmethod
from asyncio import StreamReader

from coredis._unpacker import NotEnoughData, Unpacker
from coredis._utils import b
from coredis.constants import RESPDataType
from coredis.exceptions import (
    AskError,
    AuthenticationFailureError,
    AuthenticationRequiredError,
    AuthorizationError,
    BusyLoadingError,
    ClusterCrossSlotError,
    ClusterDownError,
    ConnectionError,
    ExecAbortError,
    InvalidResponse,
    MovedError,
    NoKeyError,
    NoScriptError,
    ProtocolError,
    ReadOnlyError,
    RedisError,
    ResponseError,
    StreamConsumerGroupError,
    StreamDuplicateConsumerGroupError,
    TryAgainError,
    UnknownCommandError,
    WrongTypeError,
)
from coredis.protocols import ConnectionP
from coredis.typing import Dict, Optional, ResponseType, Set, Type, Union

try:
    import hiredis

    _hiredis_available = True
except ImportError:
    _hiredis_available = False

HIREDIS_AVAILABLE = _hiredis_available


class BaseParser(ABC):
    """
    Base class for parsers
    """

    EXCEPTION_CLASSES: Dict[
        str, Union[Type[RedisError], Dict[str, Type[RedisError]]]
    ] = {
        "ASK": AskError,
        "BUSYGROUP": StreamDuplicateConsumerGroupError,
        "CLUSTERDOWN": ClusterDownError,
        "CROSSSLOT": ClusterCrossSlotError,
        "ERR": {
            "max number of clients reached": ConnectionError,
            "unknown command": UnknownCommandError,
        },
        "EXECABORT": ExecAbortError,
        "LOADING": BusyLoadingError,
        "NOSCRIPT": NoScriptError,
        "MOVED": MovedError,
        "NOAUTH": AuthenticationRequiredError,
        "NOGROUP": StreamConsumerGroupError,
        "NOKEY": NoKeyError,
        "NOPERM": AuthorizationError,
        "NOPROTO": ProtocolError,
        "READONLY": ReadOnlyError,
        "TRYAGAIN": TryAgainError,
        "WRONGPASS": AuthenticationFailureError,
        "WRONGTYPE": WrongTypeError,
    }

    def __init__(self, read_size: int) -> None:
        self._read_size = read_size

    def parse_error(self, response: str) -> RedisError:
        """
        Parse an error response

        :meta private:
        """
        error_code = response.split(" ")[0]
        if error_code in self.EXCEPTION_CLASSES:
            response = response[len(error_code) + 1 :]
            exception_class = self.EXCEPTION_CLASSES[error_code]

            if isinstance(exception_class, dict):
                options = exception_class.items()
                exception_class = ResponseError
                for err, exc in options:
                    if response.startswith(err):
                        exception_class = exc
                        break
            return exception_class(response)
        return ResponseError(response)

    @abstractmethod
    async def read_response(
        self,
        decode: Optional[bool] = None,
        push_message_types: Optional[Set[bytes]] = None,
    ) -> ResponseType:
        pass

    @abstractmethod
    def can_read(self) -> bool:
        pass

    @abstractmethod
    def on_connect(self, connection: ConnectionP) -> None:
        """Called when the stream connects"""

    @abstractmethod
    def on_disconnect(self) -> None:
        """Called when the stream disconnects"""


class PythonParser(BaseParser):
    """
    Built in python parser that requires no additional
    dependencies.
    """

    #: Supported response data types
    def __init__(self, read_size: int) -> None:
        self._stream: Optional[StreamReader] = None
        self.encoding: Optional[str] = None
        self.unpacker: Optional[Unpacker] = None
        self.push_messages: Optional[asyncio.Queue[ResponseType]] = None

        super().__init__(read_size)

    def __del__(self) -> None:
        self.on_disconnect()

    def on_connect(self, connection: ConnectionP) -> None:
        """Called when the stream connects"""
        self._stream = connection.reader
        self.unpacker = Unpacker(connection.encoding, self.parse_error)
        self.push_messages = connection.push_messages
        if connection.decode_responses:
            self.encoding = connection.encoding

    def on_disconnect(self) -> None:
        """Called when the stream disconnects"""

        if self._stream is not None:
            self._stream = None

        if self.unpacker is not None:
            self.unpacker = None
        self.push_messages = None
        self.encoding = None

    def can_read(self) -> bool:
        return (
            (self.unpacker.bytes_written - self.unpacker.bytes_read) > 0
            if self.unpacker
            else False
        )

    async def read_response(
        self,
        decode: Optional[bool] = None,
        push_message_types: Optional[Set[bytes]] = None,
    ) -> ResponseType:
        if not self._stream:
            raise ConnectionError("Socket closed on remote end")

        need_decode = self.encoding is not None
        decode_bytes = False
        if decode is not None:
            need_decode = decode
        if need_decode and self.encoding:
            decode_bytes = True

        assert self.unpacker

        while True:
            response = self.unpacker.parse(decode_bytes)
            if isinstance(response, NotEnoughData):
                try:
                    buffer = await self._stream.read(self._read_size)
                except asyncio.CancelledError:
                    raise
                except Exception:  # noqa
                    e = sys.exc_info()[1]
                    if e:
                        raise ConnectionError(
                            f"Error {type(e)} while reading from stream: {e.args}"
                        )
                    raise
                if not buffer:
                    raise ConnectionError("Socket closed on remote end")
                self.unpacker.feed(buffer)
            else:
                if response and response.response_type == RESPDataType.PUSH:
                    assert isinstance(response.response, list)
                    assert self.push_messages
                    if (
                        not push_message_types
                        or b(response.response[0]) not in push_message_types
                    ):
                        self.push_messages.put_nowait(response.response)
                        continue
                break
        return response.response if response else None


HIREDIS_SENTINEL = False


class HiredisParser(BaseParser):
    """
    Parser class for connections using Hiredis
    """

    def __init__(self, read_size: int) -> None:
        if not HIREDIS_AVAILABLE:
            raise RuntimeError()
        self._stream: Optional[StreamReader] = None
        self._reader: Optional[hiredis.Reader] = None
        self._raw_reader: Optional[hiredis.Reader] = None
        self._next_response: Optional[
            Union[
                str,
                bytes,
                bool,  # to allow for HIREDIS_SENTINEL
            ]
        ] = HIREDIS_SENTINEL
        super().__init__(read_size)

    def __del__(self) -> None:
        self.on_disconnect()

    def can_read(self) -> bool:
        if not self._reader:  # noqa
            raise ConnectionError("Socket closed on remote end")

        if self._next_response is not HIREDIS_SENTINEL:
            self._next_response = self._reader.gets()

        return self._next_response is not HIREDIS_SENTINEL

    def on_connect(self, connection: ConnectionP) -> None:
        self._stream = connection.reader
        kwargs: Dict[str, Union[Type[Exception], str]] = {
            "protocolError": InvalidResponse,
            "replyError": ResponseError,
            # TODO: uncomment this and change HIREDIS_SENTINEL to something else
            #  so that bool responses with RESP3 can be handled.
            # "notEnoughData": HIREDIS_SENTINEL,
        }

        self._raw_reader = hiredis.Reader(**kwargs)  # type: ignore
        if connection.decode_responses:
            kwargs["encoding"] = connection.encoding
        self._reader = hiredis.Reader(**kwargs)  # type: ignore
        self._next_response = HIREDIS_SENTINEL

    def on_disconnect(self) -> None:
        if self._stream is not None:
            self._stream = None
        self._reader = None
        self._next_response = HIREDIS_SENTINEL

    async def read_response(
        self,
        decode: Optional[bool] = None,
        push_message_types: Optional[Set[bytes]] = None,
    ) -> ResponseType:
        """
        Parse a response if available on the wire

        :param decode: Only valuable if set to False to override any decode
         configuration set on the parent connection
        :return: a parsed response structure from the server
        """
        response: ResponseType
        if not self._stream:
            raise ConnectionError("Socket closed on remote end")

        # _next_response might be cached from a can_read() call
        if self._next_response is not HIREDIS_SENTINEL:
            response = self._next_response
            self._next_response = HIREDIS_SENTINEL
            return response

        cur_reader = self._reader if decode is not False else self._raw_reader
        assert cur_reader

        response = cur_reader.gets()
        while response is HIREDIS_SENTINEL:
            try:
                buffer = await self._stream.read(self._read_size)
            # CancelledError will be caught by client so that command won't be retried again
            # For more detailed discussion please see https://github.com/alisaifee/coredis/issues/56
            except asyncio.CancelledError:
                raise
            except Exception:  # noqa
                e = sys.exc_info()[1]
                if e:
                    raise ConnectionError(
                        f"Error {type(e)} while reading from stream: {e.args}"
                    )
                raise

            if not buffer:
                raise ConnectionError("Socket closed on remote end")
            cur_reader.feed(buffer)
            response = cur_reader.gets()
        if isinstance(response, ResponseError):
            return self.parse_error(response.args[0])
        return response


DefaultParser: Type[BaseParser]

if HIREDIS_AVAILABLE:
    DefaultParser = HiredisParser
else:
    DefaultParser = PythonParser



from __future__ import annotations

import datetime
import time
import warnings
from typing import TYPE_CHECKING, Optional

from packaging import version

from coredis.commands.constants import CommandName
from coredis.exceptions import CommandNotSupportedError
from coredis.typing import Union

if TYPE_CHECKING:
    import coredis.client


def normalized_seconds(value: Union[int, datetime.timedelta]) -> int:
    if isinstance(value, datetime.timedelta):
        value = value.seconds + value.days * 24 * 3600

    return value


def normalized_milliseconds(value: Union[int, datetime.timedelta]) -> int:
    if isinstance(value, datetime.timedelta):
        ms = int(value.microseconds / 1000)
        value = (value.seconds + value.days * 24 * 3600) * 1000 + ms

    return value


def normalized_time_seconds(value: Union[int, datetime.datetime]) -> int:
    if isinstance(value, datetime.datetime):
        s = int(value.microsecond / 1000000)
        value = int(time.mktime(value.timetuple())) + s

    return value


def normalized_time_milliseconds(value: Union[int, datetime.datetime]) -> int:
    if isinstance(value, datetime.datetime):
        ms = int(value.microsecond / 1000)
        value = int(time.mktime(value.timetuple())) * 1000 + ms

    return value


def check_version(
    instance: coredis.client.RedisConnection,
    command: bytes,
    function_name: str,
    min_version: Optional[version.Version],
    deprecated_version: Optional[version.Version],
    deprecation_reason: Optional[str],
) -> None:
    if not any([min_version, deprecated_version]):
        return

    if getattr(instance, "verify_version", False):
        server_version = getattr(instance, "server_version", None)
        if not server_version:
            return
        if min_version and server_version < min_version:
            raise CommandNotSupportedError(
                command.decode("latin-1"), str(instance.server_version)
            )
        if deprecated_version and server_version >= deprecated_version:
            if deprecation_reason:
                warnings.warn(
                    deprecation_reason.strip(),
                    category=DeprecationWarning,
                    stacklevel=3,
                )
            else:
                warnings.warn(
                    f"{function_name}() is deprecated since redis version {deprecated_version}. ",
                    category=DeprecationWarning,
                    stacklevel=3,
                )


def redis_command_link(command: CommandName) -> str:
    return f'`{str(command)} <https://redis.io/commands/{str(command).lower().replace(" ", "-")}>`_'



from __future__ import annotations

import functools
import inspect
from typing import Any

from coredis.exceptions import CommandSyntaxError
from coredis.typing import (
    Callable,
    Coroutine,
    Iterable,
    Optional,
    ParamSpec,
    Set,
    TypeVar,
    Union,
)

R = TypeVar("R")
P = ParamSpec("P")


class MutuallyExclusiveParametersError(CommandSyntaxError):
    def __init__(self, arguments: Set[str], details: Optional[str]):
        message = (
            f"The [{','.join(arguments)}] parameters are mutually exclusive."
            f"{' '+details if details else ''}"
        )
        super().__init__(arguments, message)


class MutuallyInclusiveParametersMissing(CommandSyntaxError):
    def __init__(self, arguments: Set[str], leaders: Set[str], details: Optional[str]):

        if leaders:
            message = (
                f"The [{','.join(arguments)}] parameters(s)"
                " must be provided together with [{','.join(leaders)}]."
                f"{' ' + details if details else ''}"
            )
        else:
            message = (
                f"The [{','.join(arguments)}] parameters are mutually"
                " inclusive and must be provided together."
                f"{' '+details if details else ''}"
            )
        super().__init__(arguments, message)


def mutually_exclusive_parameters(
    *exclusive_params: Union[str, Iterable[str]], details: Optional[str] = None
) -> Callable[
    [Callable[P, Coroutine[Any, Any, R]]], Callable[P, Coroutine[Any, Any, R]]
]:

    primary = {k for k in exclusive_params if isinstance(k, str)}
    secondary = [k for k in set(exclusive_params) - primary]

    def wrapper(
        func: Callable[P, Coroutine[Any, Any, R]]
    ) -> Callable[P, Coroutine[Any, Any, R]]:
        sig = inspect.signature(func)

        @functools.wraps(func)
        async def wrapped(*args: P.args, **kwargs: P.kwargs) -> R:
            call_args = sig.bind_partial(*args, **kwargs)
            params = {
                k
                for k in primary
                if not call_args.arguments.get(k)
                == getattr(sig.parameters.get(k), "default")
            }

            if params:
                for group in secondary:
                    for k in group:
                        if not call_args.arguments.get(k) == getattr(
                            sig.parameters.get(k), "default"
                        ):
                            params.add(k)

                            break

            if len(params) > 1:
                raise MutuallyExclusiveParametersError(params, details)

            return await func(*args, **kwargs)

        return wrapped

    return wrapper


def mutually_inclusive_parameters(
    *inclusive_params: str,
    leaders: Optional[Iterable[str]] = None,
    details: Optional[str] = None,
) -> Callable[
    [Callable[P, Coroutine[Any, Any, R]]], Callable[P, Coroutine[Any, Any, R]]
]:
    _leaders = set(leaders or [])
    _inclusive_params = set(inclusive_params)

    def wrapper(
        func: Callable[P, Coroutine[Any, Any, R]]
    ) -> Callable[P, Coroutine[Any, Any, R]]:
        sig = inspect.signature(func)

        @functools.wraps(func)
        async def wrapped(*args: P.args, **kwargs: P.kwargs) -> R:
            call_args = sig.bind_partial(*args, **kwargs)
            params = {
                k
                for k in _inclusive_params | _leaders
                if not call_args.arguments.get(k)
                == getattr(sig.parameters.get(k), "default")
            }
            if _leaders and _leaders & params != _leaders and len(params) > 0:
                raise MutuallyInclusiveParametersMissing(
                    _inclusive_params, _leaders, details
                )
            elif not _leaders and params and len(params) != len(_inclusive_params):
                raise MutuallyInclusiveParametersMissing(
                    _inclusive_params, _leaders, details
                )
            return await func(*args, **kwargs)

        return wrapped

    return wrapper



from __future__ import annotations

from io import BytesIO
from typing import NamedTuple, Optional

from coredis.constants import SYM_CRLF, RESPDataType
from coredis.exceptions import ConnectionError, InvalidResponse, RedisError
from coredis.typing import (
    Callable,
    Dict,
    Final,
    List,
    MutableSet,
    ResponsePrimitive,
    ResponseType,
    Set,
    Tuple,
    Union,
)


class RESPNode:
    __slots__ = ("container", "depth", "key", "node_type")
    depth: int
    node_type: int
    key: ResponsePrimitive

    def __init__(
        self,
        container: Union[
            List[ResponseType],
            MutableSet[Union[ResponsePrimitive, Tuple[ResponsePrimitive, ...]]],
            Dict[ResponsePrimitive, ResponseType],
        ],
        depth: int,
        node_type: int,
        key: ResponsePrimitive,
    ):
        self.container = container
        self.depth = depth
        self.node_type = node_type
        self.key = key


class NotEnoughData:
    pass


class UnpackedResponse(NamedTuple):
    response_type: int
    response: ResponseType


NOT_ENOUGH_DATA: Final[NotEnoughData] = NotEnoughData()


class Unpacker:
    ALLOWED_TYPES: Final[Set[int]] = {
        RESPDataType.NONE,
        RESPDataType.SIMPLE_STRING,
        RESPDataType.BULK_STRING,
        RESPDataType.VERBATIM,
        RESPDataType.BOOLEAN,
        RESPDataType.INT,
        RESPDataType.DOUBLE,
        RESPDataType.ARRAY,
        RESPDataType.PUSH,
        RESPDataType.MAP,
        RESPDataType.SET,
        RESPDataType.ERROR,
    }

    def __init__(
        self,
        encoding: str,
        parse_error: Callable[[str], RedisError],
    ):
        self.encoding = encoding
        self.localbuffer = BytesIO(b"")
        self.bytes_read = 0
        self.bytes_written = 0
        self.parse_error = parse_error
        self.nodes: List[RESPNode] = []

    def feed(self, data: bytes) -> None:
        self.localbuffer.seek(self.bytes_written)
        self.localbuffer.write(data)
        self.bytes_written += len(data)
        self.localbuffer.seek(self.bytes_read)

    def parse(
        self,
        decode_bytes: bool,
    ) -> Union[Optional[UnpackedResponse], NotEnoughData]:
        parsed: Optional[UnpackedResponse] = None

        assert self.localbuffer

        while True:
            data = self.localbuffer.readline()
            if not data[-2::] == SYM_CRLF:
                return NOT_ENOUGH_DATA
            self.bytes_read += len(data)
            chunk = data[:-2]

            marker, chunk = chunk[0], chunk[1:]
            response: ResponsePrimitive | RedisError = None
            if marker not in Unpacker.ALLOWED_TYPES:  # noqa
                raise InvalidResponse(
                    f"Protocol Error: {chr(marker)}, {bytes(chunk)!r}"
                )
            if marker == RESPDataType.ERROR:
                error = self.parse_error(bytes(chunk).decode())
                if isinstance(error, ConnectionError):
                    raise error
                response = error
            elif marker == RESPDataType.SIMPLE_STRING:
                response = bytes(chunk)
                if decode_bytes and self.encoding:
                    response = response.decode(self.encoding)
            elif marker == RESPDataType.INT:
                response = int(chunk)
            elif marker == RESPDataType.DOUBLE:
                response = float(chunk)
            elif marker == RESPDataType.NONE:
                response = None
            elif marker == RESPDataType.BOOLEAN:
                response = chunk[0] == ord(b"t")
            elif marker in {RESPDataType.BULK_STRING, RESPDataType.VERBATIM}:
                length = int(chunk)
                if length == -1:
                    response = None
                else:
                    if (self.bytes_written - self.bytes_read) < length + 2:
                        self.bytes_read -= len(data)
                        return NOT_ENOUGH_DATA
                    data = self.localbuffer.read(length + 2)
                    self.bytes_read += len(data)
                    response = data[:-2]
                    if marker == RESPDataType.VERBATIM:
                        if response[:3] != b"txt":
                            raise InvalidResponse(
                                f"Unexpected verbatim string of type {response[:3]!r}"
                            )
                        response = response[4:]
                    if decode_bytes and self.encoding:
                        response = response.decode(self.encoding)
            elif marker in {
                RESPDataType.ARRAY,
                RESPDataType.PUSH,
                RESPDataType.MAP,
                RESPDataType.SET,
            }:
                if marker in {RESPDataType.ARRAY, RESPDataType.PUSH}:
                    length = int(chunk)
                    if length >= 0:
                        self.nodes.append(
                            RESPNode(
                                [],
                                length,
                                marker,
                                None,
                            )
                        )
                elif marker == RESPDataType.MAP:
                    length = 2 * int(chunk)
                    if length >= 0:
                        self.nodes.append(RESPNode({}, length, marker, None))
                else:
                    length = int(chunk)
                    if length >= 0:
                        self.nodes.append(RESPNode(set(), length, marker, None))
                if length > 0:
                    continue
            if self.nodes:
                if self.nodes[-1].depth > 0:
                    self.nodes[-1].depth -= 1
                    if isinstance(self.nodes[-1].container, list):
                        self.nodes[-1].container.append(response)
                    elif isinstance(self.nodes[-1].container, dict):
                        if self.nodes[-1].key is not None:
                            self.nodes[-1].container[self.nodes[-1].key] = response
                            self.nodes[-1].key = None
                        else:
                            if isinstance(response, (bool, int, float, bytes, str)):
                                self.nodes[-1].key = response
                            else:
                                raise TypeError(
                                    f"unhashable type {response} as dictionary key"
                                )
                    else:
                        if isinstance(response, (bool, int, float, bytes, str)):
                            self.nodes[-1].container.add(response)
                        else:
                            raise TypeError(f"unhashable type {response} as set member")
                    if self.nodes[-1].depth > 1:
                        continue

                while len(self.nodes) > 1 and self.nodes[-1].depth == 0:
                    self.nodes[-2].depth -= 1
                    if isinstance(self.nodes[-2].container, list):
                        self.nodes[-2].container.append(self.nodes[-1].container)
                    elif isinstance(self.nodes[-2].container, dict):
                        if self.nodes[-2].key is not None:
                            self.nodes[-2].container[self.nodes[-2].key] = self.nodes[
                                -1
                            ].container
                            self.nodes[-2].key = None
                        else:
                            raise TypeError(
                                f"unhashable type {self.nodes[-1].container} as dictionary key"
                            )
                    else:
                        raise TypeError(
                            f"unhashable type {self.nodes[-1].container} as set member"
                        )
                    self.nodes.pop()

            if len(self.nodes) == 1 and self.nodes[-1].depth == 0:
                node = self.nodes.pop()
                parsed = UnpackedResponse(node.node_type, node.container)
                break
            if not self.nodes:
                parsed = UnpackedResponse(marker, response)
                break

        if self.bytes_read == self.bytes_written:
            self.localbuffer.seek(0)
            self.localbuffer.truncate()
            self.bytes_read = self.bytes_written = 0
        return parsed



from __future__ import annotations

import dataclasses
import functools
import textwrap
from typing import Any

from packaging import version

from coredis.commands._utils import check_version, redis_command_link
from coredis.commands.constants import CommandGroup, CommandName
from coredis.nodemanager import NodeFlag
from coredis.response._callbacks import ClusterMultiNodeCallback
from coredis.typing import Callable, Coroutine, Dict, NamedTuple, Optional, P, R


class CommandDetails(NamedTuple):
    command: bytes
    group: Optional[CommandGroup]
    readonly: bool
    version_introduced: Optional[version.Version]
    version_deprecated: Optional[version.Version]
    arguments: Dict[str, Dict[str, str]]
    cluster: ClusterCommandConfig


@dataclasses.dataclass
class ClusterCommandConfig:
    enabled: bool = True
    combine: Optional[ClusterMultiNodeCallback] = None  # type: ignore
    route: Optional[NodeFlag] = None
    split: Optional[NodeFlag] = None

    @property
    def multi_node(self) -> bool:
        return (self.route or self.split) in [
            NodeFlag.ALL,
            NodeFlag.PRIMARIES,
            NodeFlag.REPLICAS,
        ]


def redis_command(
    command_name: CommandName,
    group: Optional[CommandGroup] = None,
    version_introduced: Optional[str] = None,
    version_deprecated: Optional[str] = None,
    deprecation_reason: Optional[str] = None,
    arguments: Optional[Dict[str, Dict[str, str]]] = None,
    readonly: bool = False,
    cluster: ClusterCommandConfig = ClusterCommandConfig(),
) -> Callable[
    [Callable[P, Coroutine[Any, Any, R]]], Callable[P, Coroutine[Any, Any, R]]
]:
    command_details = CommandDetails(
        command_name,
        group,
        readonly,
        version.Version(version_introduced) if version_introduced else None,
        version.Version(version_deprecated) if version_deprecated else None,
        arguments or {},
        cluster or ClusterCommandConfig(),
    )

    def wrapper(
        func: Callable[P, Coroutine[Any, Any, R]]
    ) -> Callable[P, Coroutine[Any, Any, R]]:
        @functools.wraps(func)
        async def wrapped(*args: P.args, **kwargs: P.kwargs) -> R:
            check_version(
                args[0],  # type: ignore
                command_name,
                func.__name__,
                command_details.version_introduced,
                command_details.version_deprecated,
                deprecation_reason,
            )
            return await func(*args, **kwargs)

        wrapped.__doc__ = textwrap.dedent(wrapped.__doc__ or "")
        if group:
            wrapped.__doc__ = f"""
{wrapped.__doc__}

Redis command documentation: {redis_command_link(command_name)}
"""
        if version_introduced:
            wrapped.__doc__ += f"""
New in :redis-version:`{version_introduced}`
"""
        if version_deprecated and deprecation_reason:
            wrapped.__doc__ += f"""
Deprecated in :redis-version:`{version_deprecated}`
  {deprecation_reason.strip()}
        """
        elif version_deprecated:
            wrapped.__doc__ += f"""
Deprecated in :redis-version:`{version_deprecated}`
            """

        setattr(wrapped, "__coredis_command", command_details)
        return wrapped

    return wrapper



from __future__ import annotations

import os
import warnings
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    AnyStr,
    AsyncGenerator,
    AsyncIterator,
    Awaitable,
    Callable,
    ClassVar,
    Coroutine,
    Dict,
    Final,
    Generator,
    Generic,
    Iterable,
    Iterator,
    List,
    Literal,
    Mapping,
    MutableMapping,
    MutableSequence,
    MutableSet,
    NamedTuple,
    Optional,
    Protocol,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    ValuesView,
)

from typing_extensions import (
    Deque,
    OrderedDict,
    ParamSpec,
    Self,
    TypeAlias,
    TypedDict,
    TypeGuard,
)

_runtime_checks = False
_beartype_found = False

try:
    import beartype  # pyright: reportUnusedImport=false

    if not TYPE_CHECKING:
        from beartype.typing import (  # noqa: F811
            AbstractSet,
            Deque,
            Dict,
            Iterable,
            Iterator,
            List,
            Mapping,
            MutableMapping,
            MutableSequence,
            MutableSet,
            OrderedDict,
            Sequence,
            Set,
            Tuple,
            TypedDict,
            ValuesView,
        )

    _beartype_found = True
except ImportError:  # pragma: no cover
    pass

if (
    os.environ.get("COREDIS_RUNTIME_CHECKS", "").lower() in ["1", "true", "t"]
    and not TYPE_CHECKING
):  # pragma: no cover
    if _beartype_found:
        _runtime_checks = True
    else:
        warnings.warn(
            "Runtime checks were enabled via environment variable COREDIS_RUNTIME_CHECKS"
            " but could not import beartype"
        )

RUNTIME_TYPECHECKS = _runtime_checks

if TYPE_CHECKING:
    import coredis.exceptions

P = ParamSpec("P")
T_co = TypeVar("T_co", covariant=True)
R = TypeVar("R")


def safe_beartype(func: Callable[P, R]) -> Callable[P, R]:
    if TYPE_CHECKING:
        return func
    return beartype.beartype(func) if _beartype_found else func


def add_runtime_checks(func: Callable[P, R]) -> Callable[P, R]:
    if RUNTIME_TYPECHECKS and not TYPE_CHECKING:
        return safe_beartype(func)
    return func


CommandArgList = List[Union[str, bytes, int, float]]


#: Represents the acceptable types of a redis key
KeyT = Union[str, bytes]

#: Represents the different python primitives that are accepted
#: as input parameters for commands that can be used with loosely
#: defined types. These are encoded using the configured encoding
#: before being transmitted.
ValueT = Union[str, bytes, int, float]

#: The canonical type used for input parameters that represent "strings"
#: that are transmitted to redis.
StringT = Union[str, bytes]

#: Restricted union of container types accepted as arguments to apis
#: that accept a variable number values for an argument (such as keys, values).
#: This is used instead of :class:`typing.Iterable` as the latter allows
#: :class:`str` to be passed in as valid values for :class:`Iterable[str]` or :class:`bytes`
#: to be passed in as a valid value for :class:`Iterable[bytes]` which is never the actual
#: expectation in the scope of coredis.
#: For example::
#:
#:     def length(values: Parameters[ValueT]) -> int:
#:         return len(list(values))
#:
#:     length(["1", 2, 3, 4])      # valid
#:     length({"1", 2, 3, 4})      # valid
#:     length(("1", 2, 3, 4))      # valid
#:     length({"1": 2}.keys())     # valid
#:     length({"1": 2}.values())   # valid
#:     length(map(str, range(10))) # valid
#:     length({"1": 2})            # invalid
#:     length("123")               # invalid
#:     length(b"123")              # invalid
Parameters = Union[
    List[T_co], AbstractSet[T_co], Tuple[T_co, ...], ValuesView[T_co], Iterator[T_co]
]

#: Mapping of primitives returned by redis
ResponsePrimitive: TypeAlias = Optional[Union[StringT, int, float, bool]]

#: Represents the total structure of any response for a redis
#: command.
#:
#: This should preferably be represented by a recursive definition to allow for
#: nested containers, however limitations in both static and runtime
#: type checkers (mypy & beartype) requires loosening the definition with the use of
#: :class:`typing.Any` for now.
#:
#: Ideally the representation should be::
#:
#:    ResponseType = Union[
#:          ResponsePrimitive,
#:          List["ResponseType"],
#:          Set[ResponsePrimitive],
#:          Dict[ResponsePrimitive, "ResponseType"],
#:          "coredis.exceptions.RedisError", # response errors get mapped to exceptions.
#:      ]
ResponseType = Union[
    ResponsePrimitive,
    List[Any],
    MutableSet[Union[ResponsePrimitive, Tuple[ResponsePrimitive, ...]]],
    Dict[ResponsePrimitive, Any],
    "coredis.exceptions.RedisError",
]

__all__ = [
    "AbstractSet",
    "AnyStr",
    "AsyncIterator",
    "AsyncGenerator",
    "Awaitable",
    "Callable",
    "ClassVar",
    "CommandArgList",
    "Coroutine",
    "Deque",
    "Dict",
    "Final",
    "Generic",
    "Generator",
    "Iterable",
    "Iterator",
    "KeyT",
    "List",
    "Literal",
    "Mapping",
    "MutableMapping",
    "MutableSet",
    "MutableSequence",
    "NamedTuple",
    "OrderedDict",
    "Optional",
    "Parameters",
    "ParamSpec",
    "Protocol",
    "ResponsePrimitive",
    "ResponseType",
    "Sequence",
    "Self",
    "Set",
    "StringT",
    "Tuple",
    "Type",
    "TypeGuard",
    "TypedDict",
    "TypeVar",
    "Union",
    "ValueT",
    "ValuesView",
    "TYPE_CHECKING",
    "RUNTIME_TYPECHECKS",
]



from __future__ import annotations

import datetime
import functools
import textwrap
from typing import Any

from packaging import version

from coredis._utils import CaseAndEncodingInsensitiveEnum
from coredis.client import Redis, RedisCluster
from coredis.commands import CommandMixin
from coredis.commands._utils import (
    check_version,
    normalized_milliseconds,
    normalized_time_milliseconds,
    normalized_time_seconds,
)
from coredis.commands._wrappers import ClusterCommandConfig, CommandDetails
from coredis.commands.constants import CommandGroup
from coredis.response._callbacks import (
    BoolCallback,
    BoolsCallback,
    IntCallback,
    SimpleStringCallback,
)
from coredis.typing import (
    AnyStr,
    Callable,
    CommandArgList,
    Coroutine,
    Dict,
    Iterable,
    KeyT,
    Literal,
    Optional,
    P,
    Parameters,
    R,
    StringT,
    Tuple,
    Union,
    ValueT,
)


def _keydb_command_link(command: CommandName) -> str:
    canonical_command = str(command).lower().replace(" ", "-").replace(".", "")
    return (
        f"`{str(command)} <https://docs.keydb.dev/docs/commands#{canonical_command}>`_"
    )


class CommandName(CaseAndEncodingInsensitiveEnum):
    """
    Enum for listing all keydb extension commands
    """

    BITOP = b"BITOP"
    CRON = b"KEYDB.CRON"
    EXPIREMEMBER = b"EXPIREMEMBER"
    EXPIREMEMBERAT = b"EXPIREMEMBERAT"
    PEXPIREMEMBERAT = b"PEXPIREMEMBERAT"
    HRENAME = b"KEYDB.HRENAME"
    MEXISTS = b"KEYDB.MEXISTS"
    OBJECT_LASTMODIFIED = b"OBJECT LASTMODIFIED"
    PTTL = b"PTTL"
    TTL = b"TTL"


def keydb_command(
    command_name: CommandName,
    group: Optional[CommandGroup] = None,
    version_introduced: Optional[str] = None,
    version_deprecated: Optional[str] = None,
    deprecation_reason: Optional[str] = None,
    arguments: Optional[Dict[str, Dict[str, str]]] = None,
    readonly: bool = False,
    cluster: ClusterCommandConfig = ClusterCommandConfig(),
) -> Callable[
    [Callable[P, Coroutine[Any, Any, R]]], Callable[P, Coroutine[Any, Any, R]]
]:
    command_details = CommandDetails(
        command_name,
        group,
        readonly,
        version.Version(version_introduced) if version_introduced else None,
        version.Version(version_deprecated) if version_deprecated else None,
        arguments or {},
        cluster or ClusterCommandConfig(),
    )

    def wrapper(
        func: Callable[P, Coroutine[Any, Any, R]]
    ) -> Callable[P, Coroutine[Any, Any, R]]:
        @functools.wraps(func)
        async def wrapped(*args: P.args, **kwargs: P.kwargs) -> R:
            check_version(
                args[0],  # type: ignore
                command_name,
                func.__name__,
                command_details.version_introduced,
                command_details.version_deprecated,
                deprecation_reason,
            )
            return await func(*args, **kwargs)

        wrapped.__doc__ = textwrap.dedent(wrapped.__doc__ or "")
        if group:
            wrapped.__doc__ = f"""
{wrapped.__doc__}

KeyDB command documentation: {_keydb_command_link(command_name)}
"""

        setattr(wrapped, "__coredis_command", command_details)
        return wrapped

    return wrapper


class KeyDBCommands(CommandMixin[AnyStr]):
    @keydb_command(
        CommandName.BITOP,
        CommandGroup.BITMAP,
    )
    async def bitop(
        self,
        keys: Parameters[KeyT],
        operation: StringT,
        destkey: KeyT,
        value: Optional[int] = None,
    ) -> int:
        """
        Perform a bitwise operation using :paramref:`operation` between
        :paramref:`keys` and store the result in :paramref:`destkey`.
        """
        pieces: CommandArgList = [operation, destkey, *keys]
        if value is not None:
            pieces.append(value)
        return await self.execute_command(
            CommandName.BITOP, *pieces, callback=IntCallback()
        )

    @keydb_command(
        CommandName.CRON,
        CommandGroup.SCRIPTING,
    )
    async def cron(
        self,
        name: KeyT,
        repeat: bool,
        delay: Union[int, datetime.timedelta],
        script: StringT,
        keys: Parameters[KeyT],
        args: Parameters[ValueT],
        start: Optional[Union[int, datetime.datetime]] = None,
    ) -> bool:
        """
        Schedule a LUA script to run at a specified time and/or intervals.
        To cancel the cron delete the key at :paramref:`name`

        :param name: Name of the cron which will be visible in the keyspace,
         can be searched, and deleted with DEL.
        :param repeat: If the script will run only once, or if it will be repeated
         at the specified interval provided by :paramref:`delay`
        :param delay: is an integer specified in milliseconds used as the initial delay.
         If :paramref:`repeat` is ``True``, this will also be the length of the repeating timer
         which will execute the script each time the delay elapses
         (will continue to execute indefinitely).
        :param start: unix time specified as milliseconds enforcing that the script
         should only start executing then this Unix time has been reached.
         If :paramref:`delay` is greater than zero, this delay time will need to elapse prior to the
         script executing (timer begins to elapse at start time).
         If a start time is specified, the delay will always remain in reference
         intervals to that start time.
        :param script: is the body of the LUA script to execute.
        :param keys: The keys expected by the script
        :param args: The args required by the script
        """
        pieces: CommandArgList = [name]
        if repeat:
            pieces.append(b"REPEAT")
        else:
            pieces.append(b"SINGLE")
        if start is not None:
            pieces.append(normalized_time_milliseconds(start))
        pieces.append(normalized_milliseconds(delay))
        pieces.append(script)
        _keys = list(keys)
        pieces.append(len(_keys))
        pieces.extend(keys)
        pieces.extend(args)

        return await self.execute_command(
            CommandName.CRON, *pieces, callback=SimpleStringCallback()
        )

    @keydb_command(
        CommandName.EXPIREMEMBER,
        CommandGroup.GENERIC,
    )
    async def expiremember(
        self,
        key: KeyT,
        subkey: KeyT,
        delay: int,
        unit: Optional[Literal[b"s", b"ms"]] = None,
    ) -> bool:
        """
        Set a subkey's time to live in seconds (or milliseconds)
        """
        pieces: CommandArgList = [key, subkey, delay]
        if unit:
            pieces.append(unit.lower())
        return await self.execute_command(
            CommandName.EXPIREMEMBER, *pieces, callback=BoolCallback()
        )

    @keydb_command(
        CommandName.EXPIREMEMBERAT,
        CommandGroup.GENERIC,
    )
    async def expirememberat(
        self, key: KeyT, subkey: KeyT, unix_time_seconds: Union[int, datetime.datetime]
    ) -> bool:
        """
        Set the expiration for a subkey as a UNIX timestamp
        """
        pieces: CommandArgList = [
            key,
            subkey,
            normalized_time_seconds(unix_time_seconds),
        ]
        return await self.execute_command(
            CommandName.EXPIREMEMBERAT, *pieces, callback=BoolCallback()
        )

    @keydb_command(
        CommandName.PEXPIREMEMBERAT,
        CommandGroup.GENERIC,
    )
    async def pexpirememberat(
        self,
        key: KeyT,
        subkey: KeyT,
        unix_time_milliseconds: Union[int, datetime.datetime],
    ) -> bool:
        """
        Set the expiration for a subkey as a UNIX timestamp in milliseconds
        """
        pieces: CommandArgList = [
            key,
            subkey,
            normalized_time_milliseconds(unix_time_milliseconds),
        ]
        return await self.execute_command(
            CommandName.PEXPIREMEMBERAT, *pieces, callback=BoolCallback()
        )

    @keydb_command(CommandName.HRENAME, group=CommandGroup.HASH)
    async def hrename(
        self, key: KeyT, source_field: ValueT, destination_field: ValueT
    ) -> bool:
        """
        Rename a field :paramref:`source_field` to :paramref:`destination_field`
        in hash :paramref:`key`
        """

        return await self.execute_command(
            CommandName.HRENAME,
            key,
            source_field,
            destination_field,
            callback=BoolCallback(),
        )

    @keydb_command(CommandName.MEXISTS, readonly=True, group=CommandGroup.GENERIC)
    async def mexists(self, keys: Iterable[KeyT]) -> Tuple[bool, ...]:
        """
        Returns a tuple of bools in the same order as :paramref:`keys`
        denoting whether the keys exist
        """

        return await self.execute_command(
            CommandName.MEXISTS, *keys, callback=BoolsCallback()
        )

    @keydb_command(
        CommandName.OBJECT_LASTMODIFIED, readonly=True, group=CommandGroup.GENERIC
    )
    async def object_lastmodified(self, key: KeyT) -> int:
        """
        Returns the time elapsed (in seconds) since the key was last modified.
        This differs from idletime as it is not affected by reads of a key.

        :return: The time in seconds since the last modification
        """

        return await self.execute_command(
            CommandName.OBJECT_LASTMODIFIED, key, callback=IntCallback()
        )

    @keydb_command(CommandName.PTTL, readonly=True, group=CommandGroup.GENERIC)
    async def pttl(self, key: KeyT, subkey: Optional[ValueT] = None) -> int:
        """
        Returns the number of milliseconds until the key :paramref:`key` will expire.
        If :paramref:`subkey` is provided the response will be for the subkey.

        :return: TTL in milliseconds, or a negative value in order to signal an error
        """
        pieces: CommandArgList = [key]
        if subkey is not None:
            pieces.append(subkey)

        return await self.execute_command(
            CommandName.PTTL, *pieces, callback=IntCallback()
        )

    @keydb_command(CommandName.TTL, readonly=True, group=CommandGroup.GENERIC)
    async def ttl(self, key: KeyT, subkey: Optional[ValueT] = None) -> int:
        """
        Get the time to live for a key (or subkey) in seconds

        :return: TTL in seconds, or a negative value in order to signal an error
        """

        pieces: CommandArgList = [key]
        if subkey is not None:
            pieces.append(subkey)
        return await self.execute_command(
            CommandName.TTL, *pieces, callback=IntCallback()
        )


class KeyDB(KeyDBCommands[AnyStr], Redis[AnyStr]):
    """
    Client for `KeyDB <https://keydb.dev>`__

    The client is mostly :class:`coredis.Redis` with a couple of extra
    commands specific to KeyDB.
    """


class KeyDBCluster(KeyDBCommands[AnyStr], RedisCluster[AnyStr]):
    """
    Cluster client for `KeyDB <https://keydb.dev>`__

    The client is mostly :class:`coredis.RedisCluster` with a couple of extra
    commands specific to KeyDB.
    """
